#ifndef __RELATIONINSTANCE_H__
#define __RELATIONINSTANCE_H__

#include <vector>
#include "Value.h"
#include "Expression.h"

namespace Inter4ql {

	class Value;
	/** @brief class that implements a fact
	 *
	 */
	class Fact {
	public:
		/** @brief constructor that sets fact data
		 *
		 * @param _module module name
		 * @param a_relation relation name
		 * @param a_params vector of fact parameters
		 * @param _t is fact true value
		 * @param _f is fact false value
		 */
		explicit Fact(std::string _module, std::string a_relation, std::vector<Value *> *a_params, bool _t = true, bool _f = false);

		/** @brief constructor that sets fact data
		 *
		 * @param _module module name
		 * @param a_relation relation name
		 * @param a_params Expression being a BEL body
		 * @param _t is fact true value
		 * @param _f is fact false value
		 */
		explicit Fact(std::string _module, std::string a_relation, Expression *a_params, bool _t = true, bool _f = false);

		/** @brief constructor that sets fact data
		 *
		 * @param pointer to Fact object to be copied
		 */
		explicit Fact(Fact* f);
		/** @brief destructor
		 *
		 */
		~Fact();
		/** @brief returns string with domain information
		 *
		 * @param type do we want to print fact parameters type?
		 * @return domain information packed in a std::string
		 */
		std::string print(bool type = true) const;
		/** @brief setter for fact logical value
		 *
		 * @param _v fact logical value
		 */
		void set_value(logic_type::type _v);
		/** @brief getter for fact logical value
		 *
		 * @returns fact logical value
		 */
		Inter4ql::logic_type::type get_value() const;
		/** @brief getter for fact parameters
		 *
		 * @returns fact parameters
		 */
		std::vector<Value *> *get_params() const;
		/** @brief getter for BEL body
		 *
		 * @returns BEL body
		 */
		Expression *get_bel_params() const;
		/** @brief getter for fact module name
		 *
		 * @returns fact module name
		 */
		std::string get_module() const;
		/** @brief setter for fact module name
		 *
		 * @param _m fact module name
		 */
		void set_module(std::string _m);
		/** @brief getter for fact relation name
		 *
		 * @returns fact relation name
		 */
		std::string get_relation() const;
		/** @brief setter for set of types ('in {...}' statement)
		 *
		 * @param _v vector of types
		 */
		void set_set(std::vector<Value *> *_v);
		/** @brief getter for set of types ('in {...}' statement)
		 *
		 * @returns set of types
		 */
		std::vector<Value *> *get_set();
		/** @brief
		 *
		 * @returns 4 char that points to a view in a database
		 */
		std::string get_set_string();
		/** @brief ask for being a subset query
		 *
		 * @returns type of a query
		 */
		bool is_set_query();
		/** @brief overloaded operator used to comparing facts
		 *
		 * @param r another fact
		 */
		bool operator==(Fact &r);

		/** @brief getter for names of variables
		 *
		 * @returns vector of variable names
		 */
		std::vector<std::string> get_variable_names() const;
		/** @brief exchanges variables into values
		 *
		 * @param binding maping of variable names into values that they should get
		 */
		void valuate_fact(std::map<std::string, Value*> binding);
		/** @brief setter for info whether BEL params should be deleted on destruction
		 *
		 * @param value information value
		 */
		void should_delete_bel_params(bool value);
		/** @brief setter for info whether BEL params should be deleted on destruction
		 *
		 * @param value information value
		 */
		void set_bel_free_query(bool value);
		/** @brief setter for info whether BEL params should be deleted on destruction
		 *
		 * @param value information value
		 */
		bool is_bel_free_query();
	private:
		std::string print_params(bool type = true) const;

		std::string module;
		std::string relation;
		std::vector<Value *> *params;
		Expression *bel_params;
		bool bel_free_query;
		bool delete_bel_params;
		std::vector<Value *> *set;
		Inter4ql::logic_type::type value;
	};

	bool operator<(const Fact &a, const Fact &b);
}

#endif        /* __RELATIONINSTANCE_H__ */

