#ifndef __FUNCTIONS_CC__
#define __FUNCTIONS_CC__

#include "Functions.h"

#include "Application.h"
#include "Exception.h"
#include "Logic.h"
#include "ModuleGeneric.h"
#include "Result.h"
#include "Valuation.h"

#include <algorithm>
#include <sstream>

namespace Inter4ql {

	std::string location2string(LineParser::location_type _location) {
		std::ostringstream ss;
		ss << "line: " << _location.begin.line << ", character: " << _location.begin.column;
		return ss.str();
	}

	variable_type::type string2variable_type(std::string s){
		if(s == std::string("integer")) return variable_type::INTEGER;
		if(s == std::string("real")) return variable_type::REAL;
		if(s == std::string("literal")) return variable_type::ID;
		if(s == std::string("string")) return variable_type::STRING;
		if(s == std::string("date")) return variable_type::DATE;
		if(s == std::string("date_time")) return variable_type::DATE_TIME;
		if(s == std::string("logic")) return variable_type::LOGIC;
		return variable_type::INVALID;
	}

	std::string variable_type2string(Inter4ql::variable_type::type _variable) {
		switch (_variable) {
			case Inter4ql::variable_type::DATE: return std::string("date");
			case Inter4ql::variable_type::DATE_TIME: return std::string("date_time");
			case Inter4ql::variable_type::ID: return std::string("literal");
			case Inter4ql::variable_type::INTEGER: return std::string("integer");
			case Inter4ql::variable_type::INVALID: return std::string("invalid");
			case Inter4ql::variable_type::LOGIC: return std::string("logic");
			case Inter4ql::variable_type::REAL: return std::string("real");
			case Inter4ql::variable_type::STRING: return std::string("string");
			case Inter4ql::variable_type::VARIABLE: return std::string("variable");
			case Inter4ql::variable_type::TYPE: return std::string("type");
			default: return std::string("");
		}
	}

	class Value;
	bool bind_variables(const std::vector<Value*>& query, const std::vector<Value*>& value, std::map<std::string, Value*>& binding)
	{
		if (query.size() != value.size())
			return false;

		for (int i = 0; i < query.size(); i++) {
			if (query[i]->get_type() == Inter4ql::variable_type::VARIABLE) {
				auto it = binding.find(query[i]->get_value_string());
				if (it == binding.end()) {
					binding.insert(std::make_pair(query[i]->get_value_string(), value[i]));
				} else if (*(it->second) != *value[i]) {
					// mismatch
					return false;
				} else {
					// all is fine
				}
			} else if (*query[i] != *value[i]) {
				return false;
			} else {
				// all is fine
			}
		}
		return true;
	}

	std::vector<Value*> apply_bind(const std::vector<Value*>* query, const std::map<std::string, Value*>& binding)
	{
		std::vector <Value*> result;

		for (const auto& q : *query) {
			if (q->get_type() == variable_type::VARIABLE) {
				auto f = binding.find(q->get_value_string());
				if (f != binding.end()) {
					result.push_back(f->second);
					continue;
				}
			}
			result.push_back(q);
		}
		return result;
	}

	AsExpr* parse_asExpr(std::string toParse, int& pos) {
		if (toParse[pos] == '<') {
			AsExpr* toReturn = new AsExpr();
			toReturn->type_ = 1;
			toReturn->left_ = parse_asExpr(toParse, ++pos);
			toReturn->right_ = parse_asExpr(toParse, ++pos);
			return toReturn;
		} else {
			int nearestComma = toParse.find(",", pos);
			if (nearestComma == -1) nearestComma = toParse.length() + 1;
			int nearestClose = toParse.find(">", pos);
			if (nearestClose == -1) nearestClose = toParse.length() + 1;

			int nextStop = (nearestComma < nearestClose) ? nearestComma : nearestClose;

			AsExpr* toReturn = new AsExpr(std::string(toParse, pos, nextStop - pos));
			if (toParse[nextStop] == '>') nextStop++;
			pos = nextStop;

			return toReturn;
		}
	}

	void print_constraint(std::string& s, ConstraintEntry* constraint) {
		if (constraint) {
			std::vector<QuantifierEntry*>* q = constraint->get_quantifiers();
			for (int i = 0; i < q->size(); i++) {
				s += q->at(i)->get_type() + " " + q->at(i)->get_id() + ":" + q->at(i)->get_domain() + " ";
			}

			Expression* body = constraint->get_body();
			s += "(";
			s += print_expression(body);
			s += ")";

			std::vector<Inter4ql::Value*>* values = constraint->get_values();
			s += " in {";
			for(int i = 0; i < values->size(); i++){
				s += values->at(i)->print(false);
				if(i+1 < values->size())
					s += ", ";
			}
			s += "}";
		}
	}

	std::string print_expression(Expression* expr)
	{
		switch(expr->get_type())
		{
			case expression_type::AND: {
				std::string outL = print_expression(expr->get_left_expression());
				std::string outR = print_expression(expr->get_right_expression());
				return outL + ", " + outR;
			}
				break;
			case expression_type::IMPLIES: {
				std::string outL = print_expression(expr->get_left_expression());
				std::string outR = print_expression(expr->get_right_expression());
				std::string elseExpr;
				if (expr->get_else_expression()) {
					elseExpr = " / (" + print_expression(expr->get_else_expression()) + ")";
				}
				return outL + "-> (" + outR + ")" + elseExpr;
			}
				break;
			case expression_type::OR: {
				std::string outL = print_expression(expr->get_left_expression());
				std::string outR = print_expression(expr->get_right_expression());
				return outL + "| " + outR;
			}
				break;
			case expression_type::NOT: {
				std::string outL = print_expression(expr->get_left_expression());
				return "-" + outL;
			}
				break;
			case expression_type::QUANTIFIED: {
				std::string outQE;
				print_constraint(outQE, expr->get_quantified_entry());
				return outQE;
			}
				break;
			case expression_type::TERM:
				return expr->get_term()->print();
				break;
			case expression_type::PARALLEL: {
				std::string outL = print_expression(expr->get_left_expression());
				std::string outR = print_expression(expr->get_right_expression());
				return outL + " || " + outR;
			}
				break;
			case expression_type::SEQUENTIAL: {
				std::string outL = print_expression(expr->get_left_expression());
				std::string outR = print_expression(expr->get_right_expression());
				return outL + "; " + outR;
			}
				break;
			default:
				return "";
		}
		return "";
	}

	void print_constraints(std::string& s, std::vector<Expression*>* constraints) {
		if (constraints && constraints->size()) {
			for (int index = 0; index < constraints->size(); index++) {
				s += "\t\t";
				s += print_expression(constraints->at(index));
				s += "\n";
			}
		}
	}

	 Result check_single_valuation(std::vector<QuantifierEntry*>* quantifiers, int index,
			Result* r, std::map<std::string, Value*>* valuation, std::vector<Inter4ql::Value*>* expected,
			std::vector<Domain*> *domains, std::vector<ModuleGeneric *> *modules, const ModuleGeneric* domain_module) {
		if (index == quantifiers->size()) {
			const Valuation& v = r->get_valuation();
			const std::vector<std::string>& vn = r->get_variable_names();

			Result to_return;

			for(auto itV = v.begin(); itV != v.end(); itV++) {
				Result to_return_temp;
				std::vector<Value*> values;

				bool all_variables_match = true;

				for(int i = 0; i < vn.size(); i++) {
					auto valIt = valuation->find(vn.at(i));
					// if valuation expect some value for that variable...
					if (valIt != valuation->end()) {
						if ((*valIt->second) != (*(*itV).first.at(i))) {
							all_variables_match = false;
							break;
						}
					} else {
						values.push_back(new Value(*(*itV).first.at(i)));
						to_return_temp.add_variable_name(vn.at(i));
					}
				}

				if (!all_variables_match)
					continue;

				to_return_temp.add_element(values, itV->second);

				//printf("Reached loop end with %s\n", to_return_temp.print(true, true).c_str());
				to_return.sum_Results_OP(to_return_temp, operation_type::SUM);
			}

			//printf("Reached the end! Returning with %s\n", to_return.print(true, true).c_str());

			return to_return;
		}

		QuantifierEntry* current = quantifiers->at(index);
		VariableSpace * vs = Application::app_instance->public_get_vs();

		Result result;
		bool initialized = false;
		for (std::vector<Domain*>::iterator it = domains->begin(); it != domains->end(); it++) {
			Domain* currentD = *it;
			if (currentD->get_name() == current->get_domain()) {
				for (std::vector<ModuleGeneric *>::iterator itMod = modules->begin(); itMod != modules->end(); itMod++) {
					if (domain_module && (*itMod)->get_name() != domain_module->get_name())
						continue;
					if ((*itMod)->contains_domain(currentD)) {
						// get domain elements and their logic values
						Value* v = new Value("X");
						std::vector<Value*>* vv = new std::vector<Value*>();
						vv->push_back(v);
						Fact* toAsk = new Fact((*itMod)->get_name(), currentD->get_name(), vv);
						Result domain_result = (*itMod)->ask_for_result(toAsk);
						delete toAsk;
						delete v;

						Valuation val = domain_result.get_valuation();
						for(const auto elem : val) {
							if (elem.second == logic_type::TRUE  ||
							    elem.second == logic_type::INCONSISTENT) {
								valuation->insert(std::pair<std::string, Value*>(current->get_id(), elem.first.at(0)));
								Result result_local = check_single_valuation(quantifiers, index + 1, r, valuation, expected, domains, modules, domain_module);
								
								//printf("Before domain loop %s\n", result_local.print(true, true).c_str());

								if (current->get_type() == "forall") {
									result_local.apply_operation(operation_type::IMPLICATION, elem.second);

									//printf("After IMPLICATION %s\n", result_local.print(true, true).c_str());
									//printf("Result before AND %s\n", result.print(true, true).c_str());

									if (!initialized) {
										// initialization
										result = result_local;
										initialized = true;
									} else {
										// "and" corresponds to logical MIN
										result.sum_Results_OP(result_local, operation_type::AND);
									}

									//printf("After domain loop %s\n", result.print(true, true).c_str());
								} else {
									result_local.apply_operation(operation_type::AND, elem.second);

									if (!initialized) {
										// initialization
										result = result_local;
										initialized = true;
									} else {
										// "or" corresponds to logical MAX
										result.sum_Results_OP(result_local, operation_type::OR);
									}
								}

								valuation->erase(current->get_id());
							}
						}
					}
				}
			}
		}

		return result;
	}

	bool check_constraints(Ctraints to_check, bool& disable_constraints, std::vector<Domain*> *domains, std::vector<ModuleGeneric *> *modules, const ModuleGeneric* current_module)
	{
		if (disable_constraints)
			return true;

		disable_constraints = true;
		bool debug_constraints = Application::app_instance->debug_constraints;

		if(debug_constraints) printf("**************** checking constraints from module %s\n", current_module->get_name().c_str());
		for (std::vector<Expression*>::iterator it = to_check->begin(); it != to_check->end(); it++) {
			Expression* current = *it;
			if(debug_constraints) printf("*** Next constraint - %s\n", print_expression(current).c_str());

			int rating; // not used here...
			Result r = evaluate_expression_global(current, current_module, modules, domains, rating);
			if(debug_constraints) printf("Asked and answered! %s\n", r.print(true, true).c_str());

			auto& valuation = r.get_valuation_destructive();

			bool contains_false = false;
			bool contains_true = false;
			for(auto& val : valuation) {
				if (val.second == logic_type::UNKNOWN)
					continue;
				if (val.second == logic_type::FALSE || val.second == logic_type::INCONSISTENT) {
					contains_false = true;
				}
				if (val.second == logic_type::TRUE) {
					contains_true = true;
				}
			}

			//if(!r.get_variable_names().empty() || r.get_valuation().get_first_logic_value() != logic_type::TRUE) {
			if (contains_false || !contains_true) {
				disable_constraints = false;
				if(debug_constraints) printf("Returning FALSE\n");
				return false;
			}
   	 	}

		if(debug_constraints) printf("Returning TRUE\n");

		disable_constraints = false;
		return true;
	}

	Result evaluate_expression_global(Expression* expr, const ModuleGeneric* this_module, std::vector<ModuleGeneric*>* modules, std::vector<Domain*> *domains, int& rating) {
		Result initial;
		initial.add_element({}, logic_type::TRUE);
		return evaluate_expression_global_internal(expr, this_module, modules, initial, domains, rating);
	}

	Result evaluate_expression_global_internal(Expression* expr, const ModuleGeneric* this_module, std::vector<ModuleGeneric*>* modules,
			Result untill_now, std::vector<Domain*> *domains, int& rating)
	{
		bool debug_evaluation = Application::app_instance->debug_evaluation;
		rating = 0;

		int rating1;
		int rating2;

		switch(expr->get_type())
		{
			case expression_type::AND: {
						Result res = evaluate_expression_global_internal(expr->get_left_expression(), this_module, modules, untill_now, domains, rating1);
						if (debug_evaluation) printf("Returning from evaluate_expression_global - and, left! %s\n", res.print(true, true).c_str());
						Result res2 = evaluate_expression_global_internal(expr->get_right_expression(), this_module, modules, res, domains, rating2);
						if (debug_evaluation) printf("Returning from evaluate_expression_global - and, right! %s\n", res2.print(true, true).c_str());

						res.sum_Results_OP(res2, operation_type::AND);
						if (debug_evaluation) printf("Returning from evaluate_expression_global - and! %s\n", res.print(true, true).c_str());

						rating = rating1 + rating2;
						return res;
					  }
				break;
			case expression_type::IMPLIES: {
						Result res = evaluate_expression_global_internal(expr->get_left_expression(), this_module, modules, untill_now, domains, rating1);
						Result res2 = evaluate_expression_global_internal(expr->get_right_expression(), this_module, modules, res, domains, rating2);

						res.sum_Results_OP(res2, operation_type::IMPLICATION);
						if (debug_evaluation) printf("Returning from evaluate_expression_global - impl! %s\n", res.print(true, true).c_str());

						if (res.get_variable_names().empty() && res.get_valuation().get_first_logic_value() == logic_type::TRUE)
							rating = 2;
						return res;
					  }
				break;
			case expression_type::OR: {
						Result res = evaluate_expression_global_internal(expr->get_left_expression(), this_module, modules, untill_now, domains, rating1);
						Result res2 = evaluate_expression_global_internal(expr->get_right_expression(), this_module, modules, untill_now, domains, rating2);

						res.sum_Results_OP(res2, operation_type::OR);
						if (debug_evaluation) printf("Returning from evaluate_expression_global - or! %s\n", res.print(true, true).c_str());

						rating = std::min(rating1, rating2);
						return res;
					  }
				break;
			case expression_type::NOT: {
						Result res = evaluate_expression_global_internal(expr->get_left_expression(), this_module, modules, untill_now, domains, rating1);
						res.negate();
						if (debug_evaluation) printf("Returning from evaluate_expression_global - not! %s\n", res.print(true, true).c_str());

						if (res.get_variable_names().empty() && res.get_valuation().get_first_logic_value() == logic_type::TRUE)
							rating = 2;

						// previously it was TRUE so 2 was added - we should subtract it now
						if (res.get_variable_names().empty() && res.get_valuation().get_first_logic_value() == logic_type::FALSE)
							rating = -2;
						return res;
					  }
				break;
			case expression_type::QUANTIFIED: {
				std::map<std::string, Value*>* valuation = new std::map<std::string, Value*>();

				ConstraintEntry* current = expr->get_quantified_entry();
				//printf("*** Current quantified expression - %s\n", print_expression(current->get_body()).c_str());

				if (debug_evaluation) printf("Asking inside quantified! *************************\n");
				Result r = evaluate_expression_global(current->get_body(), this_module, modules, domains, rating1);
				if (debug_evaluation) {
					printf("Asked and answered! %s\n", r.print(true, true).c_str());
					printf("End of asking inside quantified! *************************\n");
				}

				std::vector<QuantifierEntry*>* quantifiers = current->get_quantifiers();

				const ModuleGeneric* domain_module;
				// Module from which domains should be taken...
				if (this_module->get_module_type() == module_type::BELIEF) {
					// in Belief there is a world list so no single domain_module
					domain_module = nullptr;
				} else {
					domain_module = this_module;
				}

				Result result = check_single_valuation(quantifiers, 0, &r, valuation, current->get_values(), domains, modules, domain_module);
				if (debug_evaluation)
					printf("Quantified expression checked with result %s\n", result.print(true, true).c_str());

				delete valuation;

				std::vector<Inter4ql::Value*>* expected = current->get_values();
				if (expected) {
					auto& valuation = result.get_valuation_destructive();

					for(auto& val : valuation) {
						bool is_in_logic_set = false;
						for (auto it = expected->begin(); it != expected->end(); it++) {
							if ((*it)->get_value_int() == val.second) {
								is_in_logic_set = true;
								break;
							}
						}

						if(is_in_logic_set) {
							val.second = logic_type::TRUE;
						} else {
							val.second = logic_type::FALSE;
						}
					}
				}

				if (debug_evaluation)
					printf("Returning from evaluate_expression_global - q-expression! %s\n", result.print(true, true).c_str());

				if (result.get_variable_names().empty() && result.get_valuation().get_first_logic_value() == logic_type::TRUE)
					rating = 2;

				return result;
					  }
				break;
			case expression_type::TERM: {
				if (debug_evaluation)
					printf("Evaluating term %s\n", expr->get_term()->print(true).c_str());

				std::map<logic_type::type, bool> in;
				if (expr->get_term()->is_set_query())
				{
					std::string s = expr->get_term()->get_set_string();
					in.insert(std::make_pair(logic_type::UNKNOWN      , s[0] == '1'));
					in.insert(std::make_pair(logic_type::FALSE        , s[1] == '1'));
					in.insert(std::make_pair(logic_type::TRUE         , s[2] == '1'));
					in.insert(std::make_pair(logic_type::INCONSISTENT , s[3] == '1'));
				}

				if (expr->get_term()->get_module() == "") {
					if(!this_module)
						throw new Exception("Empty module reference found in the query! This is not permitted.\n");
					Result toReturn = this_module->ask_for_result(expr->get_term());
					if (expr->get_term()->is_set_query())
					{
						for (auto& t : toReturn.get_valuation_destructive())
						{
							if (in.find(t.second)->second)
								t.second = logic_type::TRUE;
							else
								t.second = logic_type::FALSE;
						}
					}
					if (debug_evaluation)
						printf("Returning from evaluate_expression_global - term! %s\n", toReturn.print(true, true).c_str());

					if (toReturn.get_variable_names().empty() && toReturn.get_valuation().get_first_logic_value() == logic_type::TRUE)
						rating = 2;
					return toReturn;
				}
				if (this_module && expr->get_term()->get_module() == this_module->get_name()) {
					Result toReturn = this_module->ask_for_result(expr->get_term());
					if (expr->get_term()->is_set_query())
					{
						for (auto& t : toReturn.get_valuation_destructive())
						{
							if (in.find(t.second)->second)
								t.second = logic_type::TRUE;
							else
								t.second = logic_type::FALSE;
						}
					}
					if (debug_evaluation)
						printf("Returning from evaluate_expression_global - term! %s\n", toReturn.print(true, true).c_str());

					if (toReturn.get_variable_names().empty() && toReturn.get_valuation().get_first_logic_value() == logic_type::TRUE)
						rating = 2;
					return toReturn;
				}
				for (std::vector<ModuleGeneric *>::iterator itMod = modules->begin(); itMod != modules->end(); itMod++) {
					if ((*itMod)->get_name() == expr->get_term()->get_module()) {
						if (debug_evaluation)
							printf("Found matching module %s\n", (*itMod)->get_name().c_str());
						Result term_result;
						if ((*itMod)->get_module_type() == module_type::INTERNAL || (*itMod)->get_module_type() == module_type::EXTERNAL)
						{
							//printf(" and it is internal/external!\n");
							Result* till_now = untill_now.trim_variables(expr->get_term()->get_variable_names());
							//printf("Untill now: %s\n", untill_now.print(true, true).c_str());
							//printf("Till now: %s\n", till_now->print(true, true).c_str());

							for (const auto& elem : till_now->get_valuation())
							{
								if (elem.second == logic_type::UNKNOWN || elem.second == logic_type::FALSE)
									continue;
								//printf("term_result before sum: %s\n", term_result.print(true, true).c_str());

								Fact copy_of_term (expr->get_term());
								// make map from elem & till_now variables
								std::map<std::string, Value*> binding;
								for (int i = 0; i < till_now->get_variable_names().size(); i++)
								{
									binding.insert(make_pair(till_now->get_variable_names()[i], elem.first[i]));
								}
								copy_of_term.valuate_fact(binding);
								//printf("term after valuation: %s\n", copy_of_term.print(true).c_str());
								Result temp_res = (*itMod)->ask_for_result(&copy_of_term);

								// we have to pack it into proper result with proper variable assign
								Result toCheck(expr->get_term()->get_variable_names());
								std::vector<Value*> assigned_values;
								for(auto var : expr->get_term()->get_variable_names()) {
									assigned_values.push_back(binding[var]);
								}
								toCheck.add_element(assigned_values, temp_res.get_valuation().get_first_logic_value());

								term_result.sum_Results_OP(toCheck, operation_type::SUM);
								//printf("term_result after sum: %s\n", term_result.print(true, true).c_str());
							}
						} else {
							if (debug_evaluation)
								printf(" and it is NOT intermal/external!\n");
							term_result = (*itMod)->ask_for_result(expr->get_term());
							if (debug_evaluation)
								printf("term_result returned from ask_for_result: %s\n", term_result.print(true, true).c_str());
						}
						if (expr->get_term()->is_set_query())
						{
							for (auto& t : term_result.get_valuation_destructive())
							{
								if (in.find(t.second)->second)
									t.second = logic_type::TRUE;
								else
									t.second = logic_type::FALSE;
							}
						}
						if (debug_evaluation)
							printf("Returning from evaluate_expression_global - term! %s\n", term_result.print(true, true).c_str());

						if (term_result.get_variable_names().empty() && term_result.get_valuation().get_first_logic_value() == logic_type::TRUE)
							rating = 2;
						return term_result;
					}
				}
						    }
				break;
			default:
				return Result();
		}
		return Result();
	}

	database* make_snapshot(database* db) {
		database* to_return = new database();

		for(const auto& elem : *db) {
			std::string key = elem.first;

			std::shared_ptr<Valuation> val = elem.second;

			std::shared_ptr<Valuation> new_valuation = std::make_shared<Valuation>();

			for(const auto& elem : *val) {
				std::vector<Value*> new_vector;
				for(const auto& value : elem.first) {
					new_vector.push_back(new Value(value));
				}

				new_valuation->add_element(new_vector, elem.second);
			}

			to_return->insert({key, new_valuation});
		}

		return to_return;
	}

	void add_to_database(database* database, Fact* fact, Inter4ql::logic_type::type value) {
		// insert into database
		auto slot = database->find(fact->get_relation());

		if (database->find(fact->get_relation()) == database->end()) {
			database->insert(std::make_pair(fact->get_relation(), std::make_shared<Valuation>()));
			slot = database->find(fact->get_relation());
		}

		slot->second->add_element(*fact->get_params(), value);
	}

	void remove_from_database(database* database, Fact* fact) {
		// remove from database
		auto slot = database->find(fact->get_relation());

		if (database->find(fact->get_relation()) != database->end())
			slot->second->remove_element(*fact->get_params());
	}

	logic_type::type value_in_database(database* database, std::string relation, const std::vector<Value*>& vect) {
		auto slot = database->find(relation);

		if (slot == database->end())
			return logic_type::UNKNOWN;

		return slot->second->get_logic_value(vect);
	}

	bool ask_for_fact_in_database(const database* db, Fact* term, Result& res, const database* db_to_rm, bool check_existance) {
		auto variable_names = term->get_variable_names();

		auto vals = db->find(term->get_relation());
		if (vals == db->end()) {
			if (!check_existance) {
				return false;
			}
			throw new Exception("Relation " + term->get_relation() + " not found in database.\n");
		}
		auto knowledge_base = *vals->second;

		bool valuation_added = false;
		for (const auto& val : knowledge_base)
		{
			const std::vector<Value *>& knownValue = val.first;
			std::map <std::string, Value*> binding;
			bool match_found = bind_variables(*term->get_params(), knownValue, binding);

			if (!match_found)
				continue;

			std::vector<Value*> to_add;
			for (const auto& var_name: variable_names)
			{
				auto it = binding.find(var_name);
				//if (it == binding.end()) // should be impossible
				to_add.push_back(it->second);
			}

			// let's see if this element should be removed
			bool should_ignore = false;
			if (db_to_rm) {
				Fact* term_copy = new Fact(term);
				term_copy->valuate_fact(binding);

				//printf("(planning) Fact to remove - checking if remove %s\n", term_copy->print().c_str());

				auto rm_pos = db_to_rm->find(term->get_relation());
				if (rm_pos != db_to_rm->end()) {
					auto knowledge_base_rm = *rm_pos->second;

					for (const auto& val_rm : knowledge_base_rm)
					{
						const std::vector<Value *>& knownValue_rm = val_rm.first;
						std::map <std::string, Value*> binding_rm;
						bool match_found = bind_variables(*term_copy->get_params(), knownValue_rm, binding_rm);

//						printf("(planning) Comparing (");
//						for (auto val : *term_copy->get_params()) {
//							printf("%s, ", val->print().c_str());
//						}
//						printf(") with (");
//						for (auto val : knownValue_rm) {
//							printf("%s, ", val->print().c_str());
//						}
//						printf("); result %d\n", match_found);


						if (match_found) {
							should_ignore = true;
							break;
						}
					}
				}
			}

			if (should_ignore) {
				//printf("(planning) Found fact to be removed! %s\n", term->print().c_str());
				res.add_element(to_add, Inter4ql::logic_type::UNKNOWN);
			} else {
				res.add_element(to_add, val.second);
			}

			valuation_added = true;
		}

		return valuation_added;
	}

}
#endif
