#ifndef __VALUE_H__
#define __VALUE_H__

#include <string>
#include "Types.h"

namespace Inter4ql {

	/** @brief Value class that handles different types of data
	 */
	class Value {
	public:
		/** @brief constructor with constructed string value
		 *
		 * @param s value with a type string
		 */
		explicit Value(std::string s);
		/** @brief copy constructor
		 *
		 * @param value object to be copied
		 */
		explicit Value(Value* value);
		/** @brief constructor with constructed integer value
		 *
		 * @param i value with an integer type
		 * @param _type type of a value
		 */
		explicit Value(int i, Inter4ql::variable_type::type _type = Inter4ql::variable_type::INTEGER);
		/** @brief constructor with constructed double value
		 *
		 * @param d value with a double type
		 */
		explicit Value(double d);
		/** @brief destructor
		 *
		 */
		~Value();
		/** @brief setter for string type value
		 *
		 * @param s value with a string type
		 */
		void set_value(std::string s);
		/** @brief setter for integer type value
		 *
		 * @param i value with an integer type
		 * @param _type type of a value (if we want to hardtype the value)
		 */
		void set_value(int i, Inter4ql::variable_type::type _type = Inter4ql::variable_type::INTEGER);
		/** @brief setter for double type value
		 *
		 * @param d value with a double type
		 */
		void set_value(double d);
		/** @brief getter for a string type value
		 *
		 * @return value with a string type
		 */
		std::string get_value_string() const;
		/** @brief getter for a double type value
		 *
		 * @return value with a double type
		 */
		double get_value_double() const;
		/** @brief getter for an integer type value
		 *
		 * @return value with an integer type
		 */
		int get_value_int() const;
		/** @brief checks if value is numeric
		 *
		 * @return is value numeric?
		 */
		bool is_numeric();
		/** @brief gives numeric value
		 *
		 * @returns double with numeric value of this instance
		 */
		double get_value_numeric();
		/** @brief getter for a value type
		 *
		 * @return enumeration type (ID, VARIABLE, STRING, INT, DOUBLE)
		 */
		Inter4ql::variable_type::type get_type() const;
		/** @brief string comparison
		 */
		bool operator==(std::string &s);
		/** @brief integer comparison
		 */
		bool operator==(int &i);
		/** @brief double comparison
		 */
		bool operator==(double &d);
		/** @brief values comparison
		 */
		bool operator==(Value &_v);
		/** @brief values comparison
		 */
		bool operator!=(Value &_v);
		/** @brief returns string with value information
		 *
		 * @param type do we want to print information about value type?
		 * @return value information packed in a std::string
		 */
		std::string print(bool type = true);
	private:
		void set_type(Inter4ql::variable_type::type _type);

		std::string value_string;
		double value_double;
		int value_int;
		Inter4ql::variable_type::type type;
	};

	bool operator<(const Value &a, const Value &b);
	bool operator==(const Value &a, const Value &b);

}

#endif  /* __VALUE_H__ */

